'''
An implementation of the Forest-fire model described by Drossel and Schwabl
(1992). This implementation uses a 2-dimensional regular grid of cells where
each cell may be in one of three states: (1) empty, (2) occupied by a tree,
(3) burning. The state of each cell of the grid evolves over discrete time
steps according to four rules:

1. a burning cell becomes an empty cell
2. a cell occupied by a tree becomes burning if at least one cell adjacent
to the tree is burning
3. a cell occupied by a tree becomes burning with some probability even if
it has no burning cells adjacent to it
4. an empty cell becomes occupied by a tree with some probability

All cells are updated simultaneously at each time step.

Drossel B. and Schwabl F. (1992). Self-organized critical forest-fire model.
Physical Review Letters. Americal Physical Society. 69(11), 1629-1632.
'''

from state import state

'''
The probability that a tree ignites (becomes burning) during one time step.
'''
PROB_IGNITION = 0.000001

'''
The probability that a tree grows in an empty cell during one time step.
'''
PROB_GROWTH = 0.0001


def rows_in(cells):
    '''
    Returns the number of rows in a two-dimensional list of cells.

    Parameters
    ---------
    cells : list-of-list of bool
        The cells of a 2D Forest-fire.

    Returns
    -------
    int
        The number or rows in `cells`.
    '''
    return len(cells)


def cols_in(cells):
    '''
    Returns the number of columns in a two-dimensional list of cells.

    Parameters
    ---------
    cells : list-of-list of bool
        The cells of a 2D Forest-fire.

    Returns
    -------
    int
        The number or columns in `cells`.
    '''
    if rows_in(cells) > 0:
        return len(cells[0])


def index_ok(cells, row, col):
    '''
    Returns True if row and col are valid indexes for the two-dimensional
    list of cells, False otherwise.

    Parameters
    ---------
    cells : list-of-list of state
        The cells of a 2D Forest-fire
    row : int
        A row index.
    col : int
        A column index.

    Returns
    -------
    bool
        True if row and col are valid indexes for the two-dimensional
        list of cells.
    '''

    if row < 0 or row >= rows_in(cells):
        return False
    if col < 0 or col >= cols_in(cells):
        return False
    return True



def print_forest(cells):
    '''
    Prints the cells of a Forest-fire model followed by a new line.
	Empty cells are represented with a hyphen '-', trees are represented with the
	character 'T', and burning trees are represented with the character '#'.
	There are no separator characters between cells. Each row of cells is
	printed on a separate line.

    Parameters
    ----------
    cells : list-of-list of state
        The cells of a 2D Forest-fire
    '''
    rows = rows_in(cells)
    cols = cols_in(cells)
    for r in range(0, rows):
        for c in range(0, cols):
            s = cells[r][c]
            if s == state.EMPTY:
                print('-', end = '')
            elif s == state.OCCUPIED:
                print('T', end = '')
            else:
                print('#', end = '')
        print()


def copy(cells):
    '''
    Returns a new two-dimensional list that is equal to the two-dimensional list
    `cells`.

    The returned list has the same number of rows and columns as `cells`
    and each element of the returned list is equal to the corresponding
    element in `cells`.

    Parameters
    ---------
    cells : list-of-list of state
        The cells of a 2D Forest-fire.

    Returns
    -------
    list
        A list-of-list of state equal to `cells`
    '''
    rows = rows_in(cells)
    cols = cols_in(cells)

    # make a two-dimensional list having rows rows and cols columns
    # all values are set to None
    # In Java, you can do all of this in one line of code...
    #
    # start with an empty list
    copy = []
    # make a list for each row of the original list
    for i in range(0, rows):
        # each row is made up of cols columns
        row = [None] * cols
        copy.append(row)

    # copy the values from cells into copy
    for r in range(0, rows):
        for c in range(0, cols):
            copy[r][c] = cells[r][c]
    return copy

    

def n_moore(cells, row, col):
    '''
    Return the Moore neighborhood of cells around a specified cell.

    The Moore neighborhood of a cell consists of cells[row][col] and its
    eight neighbours, which are the cells that are horizontally, vertically,
    or diagonally adjacent (if those neighbors exist).

    Parameters
    ----------
    cells : list-of-list of state
        The cells of a 2D Forest-fire.
    row : int
        The row index of the cell to get the neigborhood of.
    col : int
        The column index of the cell to get the neigborhood of.

    Returns
    -------
    list
        A 3x3 list-of-list of state containing the cells in the Moore
        neigborhood of a specified cell.

    Raises
    ------
    ValueError
        If `row` or `col` is not a valid non-negative index for cells
    '''
    if not index_ok(cells, row, col):
        raise ValueError()

    # make a 3x3 list-of-list to store the neighborhood
    nhood = [[None, None, None],
        [None, None, None],
        [None, None, None]]

    # upper-left indexes of the neighborhood
    # these might be out of bounds
    left = col - 1
    top = row - 1

    # get the neighborhood
    for r in range(0, 3):
        # row index for cells
        cellsRow = top + r
        for c in range(0, 3):
            # column index for cells
            cellsCol = left + c
            if index_ok(cells, cellsRow, cellsCol):
                nhood[r][c] = cells[cellsRow][cellsCol]
            else:
                nhood[r][c] = state.EMPTY
    return nhood


def is_burning(cells, row, col):
    '''
    Return True if a specified cell is burning, and False otherwise.

    Parameters
    ----------
    cells: list-of-list of state
        The cells of a 2D Forest-fire.
    row : int
        The row index of the cell.
    col : int
        The column index of the cell.

    Returns
    -------
    bool
        True if the specified cell is burning, False otherwise.

    Raises
    ------
    ValueError
        If `row` or `col` is not a valid non-negative index for cells
    '''
    if not index_ok(cells, row, col):
        raise ValueError()

    return cells[row][col] == state.BURNING


def num_burning(cells):
    '''
    Returns the total number of cells that are burning in a Forest-fire.

    Parameters
    ----------
    list-of-list of state
        The cells of a 2D Forest-fire.

    Returns
    -------
    int
        The number of burning cells in `cells`
    '''
    rows = rows_in(cells)
    cols = cols_in(cells)
    n = 0
    for r in range(0, rows):
        for c in range(0, cols):
            if is_burning(cells, r, c):
                n += 1
    return n


